#include "BSP.h"
#include <math.h>
#include <stdlib.h>

/*

	BSP tree stuff.
	
	This is for correct rendering of sections with partial transparency. It is not
	used for culling or collisions or anything else. As the main cost will be OpenGL
	calls for face drawing, the tree is constructed to have the minimum number of
	splits.

*/
CBSP::CBSPNode::CBSPNode()
{
	Back = Front = Undivided = NULL;
}

CBSP::CBSPNode::~CBSPNode()
{
	if(Back) delete Back;
	if(Front) delete Front;
}

void CBSP::CBSPNode::Draw(GLfloat *Matrix)
{
	/* use matrix to determine whether we are in front or behind this plane */
	float TransNormal[3];

	TransNormal[0] = Matrix[0]*Normal[0] + Matrix[4]*Normal[1] + Matrix[8]*Normal[2];
	TransNormal[1] = Matrix[1]*Normal[0] + Matrix[5]*Normal[1] + Matrix[9]*Normal[2];
	TransNormal[2] = Matrix[2]*Normal[0] + Matrix[6]*Normal[1] + Matrix[10]*Normal[2];
	
	float Side = -Matrix[12]*TransNormal[0] -Matrix[13]*TransNormal[1] -Matrix[14]*TransNormal[2] - Distance;
	if(Side < 0)
	{
		if(Front) Front->Draw(Matrix);
		Material->Select();
		glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_INT, Elements);
		if(Back) Back->Draw(Matrix);
	}
	else
	{
		if(Back) Back->Draw(Matrix);
		Material->Select();
		glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_INT, Elements);
		if(Front) Front->Draw(Matrix);
	}
}

void CBSP::Draw()
{
	if(Head)
	{
		glInterleavedArrays(GL_T2F_V3F, 0, Verts);

		GLfloat ModelView[16];
		glGetFloatv(GL_MODELVIEW_MATRIX, ModelView);
		Head->Draw(ModelView);
	}
}

void CBSP::Setup()
{
	DeleteGeometryHolders();
}

void CBSP::Add(GLuint *elem, int numelem, CMaterial *mat)
{
	GeometryHolder *NewGeom = new GeometryHolder;
	NewGeom->Next = GHead;
	GHead = NewGeom;
	NewGeom->Elements = elem;
	NewGeom->NumElements = numelem;
	NewGeom->Material = mat;
}

CBSP::CBSP()
{
	Head = NULL;
	GHead = NULL;
	CurrentVert = NumAllocdVerts = 0;
	Verts = NULL;
}

void CBSP::DeleteGeometryHolders()
{
	while(GHead)
	{
		delete[] GHead->Elements;
		GeometryHolder *Next = GHead->Next;
		delete GHead;
		GHead = Next;
	}
}

CBSP::~CBSP()
{
	delete Head;
	DeleteGeometryHolders();
	delete[] Verts;
}

/*

	BSP COMPILER

*/

CBSP::CBSPNode *CBSP::CBSPNode::GetDivider(CBSP *Parent)
{
	/* look for the node that will create the least splits */
	int LowSplits = GetNumDivisions(Parent, this);
	CBSPNode *LowRoller = this;
	CBSPNode *List = Undivided;
	while(List)
	{
		int NumSplits = List->GetNumDivisions(Parent, this);
		if(NumSplits < LowSplits)
		{
			LowSplits = NumSplits;
			LowRoller = List;
		}
		List = List->Undivided;
	}
	return LowRoller;
}

float Distances[3];
int NewElements[4];
int NumElements;

int CBSP::CBSPNode::GetPoint(CBSP *Parent, float *First, float *Second, float ratio)
{
	float NewPoint[5];
	int c = 5;
	while(c--)
		NewPoint[c] = (1-ratio)*First[c] + ratio*Second[c];
	return Parent->AddVert(NewPoint);
}

void CBSP::CBSPNode::GetSplitPoints(CBSP *Parent, CBSPNode *Splitee, float Side)
{
	/* create negative side(s) */
	NumElements = 0;
	for(int c = 0; c < 3; c++)
	{
		int c2 = (c+1)%3;

		if(Side*Distances[c] >= 0)
		{
			NewElements[NumElements++] = Splitee->Elements[c];
						
			if(Side*Distances[c2] < 0)
			{
				float ratio = Distances[c] / (Distances[c] - Distances[c2]);
				NewElements[NumElements++] = GetPoint(Parent, &Parent->Verts[Splitee->Elements[c]*5], &Parent->Verts[Splitee->Elements[c2]*5], ratio);
			}
		}
		else
			if(Side*Distances[c2] >= 0)
			{
				float ratio = Distances[c] / (Distances[c] - Distances[c2]);
				NewElements[NumElements++] = GetPoint(Parent, &Parent->Verts[Splitee->Elements[c]*5], &Parent->Verts[Splitee->Elements[c2]*5], ratio);
			}
	}
}


void CBSP::CBSPNode::Split(CBSP *Parent, CBSP::CBSPNode *Splitee)
{
	CBSPNode *NewNode;
	
	/* deal with positive side first */
	GetSplitPoints(Parent, Splitee, +1);

		// store 1 or 2 new faces
		NewNode = new CBSPNode;
		*NewNode = *Splitee;
		NewNode->Elements[0] = NewElements[0];
		NewNode->Elements[1] = NewElements[1];
		NewNode->Elements[2] = NewElements[2];
		NewNode->Undivided = Front;
		Front = NewNode;

		if(NumElements == 4)
		{
			NewNode = new CBSPNode;
			*NewNode = *Splitee;
			NewNode->Elements[0] = NewElements[0];
			NewNode->Elements[1] = NewElements[2];
			NewNode->Elements[2] = NewElements[3];
			NewNode->Undivided = Front;
			Front = NewNode;
		}

	/* deal with negative side second */
	GetSplitPoints(Parent, Splitee, -1);

		// store 1 or 2 new faces - reusing memory
		NewNode = new CBSPNode;
		*NewNode = *Splitee;
		NewNode->Elements[0] = NewElements[0];
		NewNode->Elements[1] = NewElements[1];
		NewNode->Elements[2] = NewElements[2];
		NewNode->Undivided = Back;
		Back = NewNode;

		if(NumElements == 4)
		{
			NewNode = new CBSPNode;
			*NewNode = *Splitee;
			NewNode->Elements[0] = NewElements[0];
			NewNode->Elements[1] = NewElements[2];
			NewNode->Elements[2] = NewElements[3];
			NewNode->Undivided = Back;
			Back = NewNode;
		}
}

void CBSP::CBSPNode::Divide(CBSP *Parent, CBSP::CBSPNode *Children)
{
	/* divide list */
	while(Children)
	{
		CBSPNode *Next = Children->Undivided;
		if(Children != this)
		{
			int side = ClassifyFace(Parent, Children);
			if(side)
			{
				if(side > 0)
				{
					Children->Undivided = Front;
					Front = Children;
				}
				else
				{
					Children->Undivided = Back;
					Back = Children;
				}
			}
			else
			{
				Split(Parent, Children);
				delete Children;
			}
		}
		Children = Next;
	}

	/* recurse */
	if(Back)
	{
		CBSPNode *B = Back->GetDivider(Parent);
		B->Divide(Parent, Back);
		Back = B;
	}
	if(Front)
	{
		CBSPNode *F = Front->GetDivider(Parent);
		F->Divide(Parent, Front);
		Front = F;
	}
}

#define SPLIT_QUANTUM 0.01f
int CBSP::CBSPNode::ClassifyFace(CBSP *Parent, CBSP::CBSPNode *Face)
{
	int c = 3;
	while(c--)
		Distances[c] =	Normal[0]*Parent->Verts[Face->Elements[c]*5 + 2] +
						Normal[1]*Parent->Verts[Face->Elements[c]*5 + 3] +
						Normal[2]*Parent->Verts[Face->Elements[c]*5 + 4] -
						Distance;
			
	if(Distances[0] <= SPLIT_QUANTUM && Distances[1] <= SPLIT_QUANTUM && Distances[2] <= SPLIT_QUANTUM) return -1;
	if(Distances[0] >= -SPLIT_QUANTUM && Distances[1] >= -SPLIT_QUANTUM && Distances[2] >= -SPLIT_QUANTUM) return +1;
	return 0;
}

int CBSP::CBSPNode::GetNumDivisions(CBSP *Parent, CBSPNode *Children)
{
	/* determine how many of Children (that aren't this node) cross this face */
	int Splits = 0;
	while(Children)
	{
		if(Children != this)
		{
			if(!ClassifyFace(Parent, Children))
				Splits++;
		}
		Children = Children->Undivided;
	}
	return Splits;
}

void CBSP::CBSPNode::ComputeNormal(CBSP *Parent)
{
	float SideVec1[3], SideVec2[3];

	/* compute two edge vectors */
	SideVec1[0] = Parent->Verts[Elements[0]*5 + 2] - Parent->Verts[Elements[1]*5 + 2];
	SideVec1[1] = Parent->Verts[Elements[0]*5 + 3] - Parent->Verts[Elements[1]*5 + 3];
	SideVec1[2] = Parent->Verts[Elements[0]*5 + 4] - Parent->Verts[Elements[1]*5 + 4];

	SideVec2[0] = Parent->Verts[Elements[1]*5 + 2] - Parent->Verts[Elements[2]*5 + 2];
	SideVec2[1] = Parent->Verts[Elements[1]*5 + 3] - Parent->Verts[Elements[2]*5 + 3];
	SideVec2[2] = Parent->Verts[Elements[1]*5 + 4] - Parent->Verts[Elements[2]*5 + 4];

	/* use cross product to get normal */
	Normal[0] = SideVec1[1]*SideVec2[2] - SideVec2[1]*SideVec1[2];
	Normal[1] = SideVec1[2]*SideVec2[0] - SideVec2[2]*SideVec1[0];
	Normal[2] = SideVec1[0]*SideVec2[1] - SideVec2[0]*SideVec1[1];

	/* normalise normal */
	float LengthMul = 1.0f / sqrt(Normal[0]*Normal[0] + Normal[1]*Normal[1] + Normal[2]*Normal[2]);
	Normal[0] *= LengthMul;
	Normal[1] *= LengthMul;
	Normal[2] *= LengthMul;

	/* calculate distance from origin */
	Distance = Parent->Verts[Elements[0]*5 + 2]*Normal[0] + Parent->Verts[Elements[0]*5 + 3]*Normal[1] + Parent->Verts[Elements[0]*5 + 4]*Normal[2];
}

void CBSP::Compile(GLfloat *Verts)
{
	int CurrentFace = 0;
	GeometryHolder *GParser = GHead;

	while(GParser)
	{
		if(CurrentFace < GParser->NumElements)
		{
			CBSPNode *Next = Head;
			Head = new CBSPNode;
			Head->Elements[0] = AddVert(&Verts[GParser->Elements[CurrentFace++]*5]);
			Head->Elements[1] = AddVert(&Verts[GParser->Elements[CurrentFace++]*5]);
			Head->Elements[2] = AddVert(&Verts[GParser->Elements[CurrentFace++]*5]);
			Head->Material = GParser->Material;
			Head->ComputeNormal(this);
			Head->Undivided = Next;
		}
		else
		{
			CurrentFace = 0;
			GParser = GParser->Next;
		}
	}
	DeleteGeometryHolders();

	/* decide where to start division */
	CBSPNode *H = Head->GetDivider(this);

	/* start division process */
	H->Divide(this, Head);
	Head = H;
}

#define SIMILARITY_QUANTUM	0.001f
#define ALLOC_SIZE	32
int CBSP::AddVert(GLfloat *input)
{
	/* search for vert in existing list */
	int snum = CurrentVert;
	while(snum--)
	{
		int c = 5;
		bool Same = true;
		while(c--)
		{
			if(fabs(Verts[snum*5 + c] - input[c]) >= SIMILARITY_QUANTUM)
			{
				Same = false;
				break;
			}
		}

		if(Same)
			return snum;
	}

	/* need to add this vert... */
	if(CurrentVert == NumAllocdVerts)
	{
		NumAllocdVerts += ALLOC_SIZE;
		Verts = (GLfloat *)realloc(Verts, sizeof(GLfloat)*NumAllocdVerts*5);
	}

//	printf("[%d]\n", CurrentVert); fflush(stdout);
	int c = 5;
	while(c--)
		Verts[CurrentVert*5 + c] = input[c];
	return CurrentVert++;
}
