#include "Object.h"
#include <lib3ds/file.h>
#include <lib3ds/material.h>
#include <lib3ds/mesh.h>
#include "Material.h"
#include <stdlib.h>
#include <string.h>
#include "NvTriStrip.h"

CObject::CObject()
{
	MeshList = NULL;
}

CObject::~CObject()
{
	while(MeshList)
	{
		Mesh *Next = MeshList->Next;
		delete[] MeshList->Verts;
		delete[] MeshList->Faces;
		delete MeshList;
		MeshList = Next;
	}
}

int FaceCompare(const void *a, const void *b)
{
	return strcmp( ((Lib3dsFace *)a)->material, ((Lib3dsFace *)b)->material);
}

bool CObject::Open(CPString &Path, CPString &Name)
{
	Lib3dsFile *file;
	Path.CompletePath();
	CPString Total = Path + Name;
	CCollection LocalMaterials;

	if(!(file = lib3ds_file_load(Total))) return false;

	/* store object name */
	*dynamic_cast<CPString *>(this) = Total;

	Lib3dsMaterial *Mat = file->materials;
	while(Mat)
	{
		CPString Name = GetMaterialName(Mat);
		CMaterial *FoundMat;
		if(!(FoundMat = dynamic_cast<CMaterial *>(MaterialCollection.Find(Name))))
		{
			FoundMat = new CMaterial;
			FoundMat->Setup(Path, Mat);
			MaterialCollection.Add(FoundMat);
		}

		/* put an entry into our local -> global name lookup */
		CPStringMapping *NewMapping = new CPStringMapping;
		NewMapping->AlternateName = Name;
		*dynamic_cast<CPString *>(NewMapping) = Mat->name;
		LocalMaterials.Add(NewMapping);

		Mat = Mat->next;
	}

	Lib3dsMesh *MeshP = file->meshes;
	while(MeshP)
	{
		Mesh *Next = MeshList;
		MeshList = new Mesh;
		MeshList->Next = Next;
		MeshList->Faces = NULL;
		
		/* seed BSP */
		MeshList->Transparent = new CBSP;
		MeshList->Transparent->Setup();

		/* copy matrix */
		unsigned int c = 16;
		while(c--)
			MeshList->Matrix[c] = MeshP->matrix[c >> 2][c&3];

		/* allocate geometry */
		MeshList->NumVerts = MeshP->points;
		MeshList->Verts = new GLfloat[MeshList->NumVerts * 5];
		
		/* copy geometry */
		c = MeshList->NumVerts;
		while(c--)
		{
			int ic = c*5;
			/* format is u, v, x, y, z */
			if(c < MeshP->texels)
			{
				MeshList->Verts[ic+0] = MeshP->texelL[c][0];
				MeshList->Verts[ic+1] = MeshP->texelL[c][1];
			}

			MeshList->Verts[ic+2] = MeshP->pointL[c].pos[0];
			MeshList->Verts[ic+3] = MeshP->pointL[c].pos[1];
			MeshList->Verts[ic+4] = MeshP->pointL[c].pos[2];
		}

		/* sort face list */
		qsort(MeshP->faceL, MeshP->faces, sizeof(Lib3dsFace), FaceCompare);
		unsigned int FaceC = 0;
		while(FaceC < MeshP->faces)
		{
			/* work out how many faces use this material */
			unsigned int EndC = FaceC;
			while(EndC < MeshP->faces && !strcmp(MeshP->faceL[EndC].material, MeshP->faceL[FaceC].material))
				EndC++;

			/* find material */
			CPString SearchString1; SearchString1 = (const char *)MeshP->faceL[FaceC].material;
			CMaterial *Material = (CMaterial *)MaterialCollection.Find((((CPStringMapping *)LocalMaterials.Find(SearchString1))->AlternateName));

			/* either get an optimal strip or add to BSP depending on transparency */
			if(Material->IsTransparent())
			{
				GLuint *Elements = new GLuint[(EndC-FaceC)*3];
				c = EndC-FaceC;
				while(c--)
				{
					Elements[(c*3)] = MeshP->faceL[FaceC+c].points[0];
					Elements[(c*3)+1] = MeshP->faceL[FaceC+c].points[1];
					Elements[(c*3)+2] = MeshP->faceL[FaceC+c].points[2];
				}

				/* add to BSP */
				MeshList->Transparent->Add(Elements, (EndC-FaceC)*3, Material);
			}
			else
			{
				/* allocate face list */
				Mesh::FaceSet *NewSet = new Mesh::FaceSet;
				NewSet->Next = MeshList->Faces;
				MeshList->Faces = NewSet;

				/* store material */
				NewSet->Material = Material;

				/* temporary face list */
				unsigned short *TFaceList = new unsigned short[(EndC-FaceC)*3];
				c = EndC-FaceC;
				while(c--)
				{
					TFaceList[(c*3)] = MeshP->faceL[FaceC+c].points[0];
					TFaceList[(c*3)+1] = MeshP->faceL[FaceC+c].points[1];
					TFaceList[(c*3)+2] = MeshP->faceL[FaceC+c].points[2];
				}
				PrimitiveGroup *OutGroup;
				unsigned short NumGroups;
				GenerateStrips(TFaceList, (EndC-FaceC)*3, &OutGroup, &NumGroups);

				/* now store things from generated strips */
				NewSet->NumFaces = OutGroup->numIndices;
				NewSet->Faces = new GLuint[NewSet->NumFaces];

				c = NewSet->NumFaces;
				while(c--)
					NewSet->Faces[c] = OutGroup->indices[c];

				switch(OutGroup->type)
				{
					case PT_LIST: NewSet->Mode = GL_TRIANGLES; break;
					case PT_STRIP: NewSet->Mode = GL_TRIANGLE_STRIP; break;
					case PT_FAN: NewSet->Mode = GL_TRIANGLE_FAN; break;
				}
				delete[] OutGroup;
			}

			FaceC = EndC;
		}

		/* compile or delete BSP as appropriate */
		if(MeshList->Transparent->Empty())
		{
			delete MeshList->Transparent;
			MeshList->Transparent = NULL;
		}
		else
			MeshList->Transparent->Compile(MeshList->Verts);

		MeshP = MeshP->next;
	}

	LocalMaterials.DeleteAll();
	lib3ds_file_free(file);
	return true;
}

void CObject::DrawTransparent()
{
	Mesh *M = MeshList;
	while(M)
	{
		if(M->Transparent)
		{
			glPushMatrix();
				glMultMatrixf(M->Matrix);
				M->Transparent->Draw();
			glPopMatrix();
		}

		M = M->Next;
	}
}

void CObject::DrawSolid()
{
	Mesh *M = MeshList;
	while(M)
	{
		if(M->Faces)
		{
			/* setup vertices */
			glInterleavedArrays(GL_T2F_V3F, 0, M->Verts);

			/* draw faces */
			glPushMatrix();

				glMultMatrixf(M->Matrix);

				Mesh::FaceSet *FS = M->Faces;
				while(FS)
				{
					FS->Material->Select();
					glDrawElements(FS->Mode, FS->NumFaces, GL_UNSIGNED_INT, FS->Faces);
					FS = FS->Next;
				}

			glPopMatrix();
		}

		M = M->Next;
	}
}
